@@ -38,14 +38,19 @@
 #include <unistd.h>
 
 #include <glib/gi18n.h>
+#include <gio/gio.h>
 #include <gdk/gdkkeysyms.h>
 #include <glade/glade-xml.h>
 #include <libgnome/gnome-exec.h>
 #include <libgnome/gnome-util.h>
 #include <libgnomeui/gnome-entry.h>
-#include <libgnomeui/gnome-url.h>
-#include <libgnomevfs/gnome-vfs-utils.h>
 #include <gconf/gconf-client.h>
+#include <gmenu-tree.h>
+
+#include <libpanel-util/panel-error.h>
+#include <libpanel-util/panel-glib.h>
+#include <libpanel-util/panel-keyfile.h>
+#include <libpanel-util/panel-show.h>
 
 #include "nothing.h"
 #include "panel-gconf.h"
@@ -58,8 +63,7 @@
 #include "menu.h"
 #include "panel-lockdown.h"
 #include "panel-xutils.h"
-
-#define ICON_SIZE 48.0
+#include "panel-icon-names.h"
 
 typedef struct {
 	GtkWidget        *run_dialog;
@@ -112,6 +116,8 @@ enum {
 
 static PanelRunDialog *static_dialog = NULL;
 
+static void panel_run_dialog_disconnect_pixmap (PanelRunDialog *dialog);
+
 static void
 panel_run_dialog_destroy (PanelRunDialog *dialog)
 {
@@ -166,6 +172,8 @@ panel_run_dialog_destroy (PanelRunDialog *dialog)
 	if (dialog->completion)
 		g_completion_free (dialog->completion);
 	dialog->completion = NULL;
+
+	panel_run_dialog_disconnect_pixmap (dialog);
 	
 	g_free (dialog);
 }
@@ -174,37 +182,49 @@ static void
 panel_run_dialog_set_default_icon (PanelRunDialog *dialog, gboolean set_drag)
 {
 	gtk_image_set_from_icon_name (GTK_IMAGE (dialog->pixmap),
-				      PANEL_RUN_ICON,
+				      PANEL_ICON_RUN,
 				      GTK_ICON_SIZE_DIALOG);
 	
 	gtk_window_set_icon_name (GTK_WINDOW (dialog->run_dialog),
-				  PANEL_RUN_ICON);
+				  PANEL_ICON_RUN);
 
 	if (set_drag)
 		gtk_drag_source_set_icon_name (dialog->run_dialog,
-					       PANEL_RUN_ICON);
+					       PANEL_ICON_LAUNCHER);
 }
 
 static void
 panel_run_dialog_set_icon (PanelRunDialog *dialog,
-			   const char     *icon_path)
+			   const char     *icon_path,
+			   gboolean        force)
 {
 	GdkPixbuf *pixbuf = NULL;
 
+	if (!force && icon_path && dialog->icon_path &&
+	    !strcmp (icon_path, dialog->icon_path))
+		return;
+
 	g_free (dialog->icon_path);
 	dialog->icon_path = NULL;
 		
-	if (icon_path && dialog->icon_path &&
-	    !strcmp (icon_path, dialog->icon_path))
-		return;
+	if (icon_path) {
+		GdkScreen   *screen;
+		GtkSettings *settings;
+		int          size;
+
+		screen = gtk_widget_get_screen (GTK_WIDGET (dialog->pixmap));
+		settings = gtk_settings_get_for_screen (screen);
+		gtk_icon_size_lookup_for_settings (settings,
+						   GTK_ICON_SIZE_DIALOG,
+						   &size, NULL);
 
-	if (icon_path)
 		pixbuf = panel_load_icon (gtk_icon_theme_get_default (),
 					  icon_path,
-					  ICON_SIZE,
-					  ICON_SIZE,
-					  ICON_SIZE,
+					  size,
+					  size,
+					  size,
 					  NULL);
+	}
 	
 	if (pixbuf) {
 		dialog->icon_path = g_strdup (icon_path);
@@ -214,6 +234,8 @@ panel_run_dialog_set_icon (PanelRunDialog *dialog,
 		 */
 		gtk_image_set_from_pixbuf (GTK_IMAGE (dialog->pixmap), pixbuf);
 
+		//FIXME: it'd be better to set an icon of the correct size,
+		//(ditto for the drag icon?)
 		gtk_window_set_icon (GTK_WINDOW (dialog->run_dialog), pixbuf);
 		
 		gtk_drag_source_set_icon_pixbuf (dialog->run_dialog, pixbuf);
@@ -231,7 +253,7 @@ command_is_executable (const char *command)
 	char      *path;
 	int        argc;
 	
-	path = gnome_vfs_get_local_path_from_uri (command);
+	path = g_filename_from_uri (command, NULL, NULL);
 	if (!path)
 		path = g_strdup (command);
 	
@@ -316,34 +338,6 @@ panel_run_dialog_launch_command (PanelRunDialog *dialog,
 	return result;
 }
 
-static gboolean
-panel_run_dialog_show_url (PanelRunDialog *dialog,
-	                   const char     *url,
-	                   const char     *escaped)
-{
-	GError    *error = NULL;
-	GdkScreen *screen;
-	
-	screen = gtk_window_get_screen (GTK_WINDOW (dialog->run_dialog));
-
-	gnome_url_show_on_screen (url, screen, &error);
-	if (error) {
-		char *primary;
-
-		primary = g_strdup_printf (_("Could not open location '%s'"),
-					   escaped);
-		panel_error_dialog (GTK_WINDOW (dialog->run_dialog), NULL,
-				    "cannot_show_url", TRUE,
-				    primary, error->message);
-		g_free (primary);
-
-		g_error_free (error);
-		return FALSE;
-	}
-	
-	return TRUE;
-}
-
 static void
 panel_run_dialog_execute (PanelRunDialog *dialog)
 {
@@ -352,7 +346,8 @@ panel_run_dialog_execute (PanelRunDialog *dialog)
 	gboolean   result;
 	char      *command;
 	char      *escaped;
-	char      *disk, *url;
+	GFile     *file;
+	char      *disk, *uri;
 	char      *scheme;	
 	
 	screen = gtk_window_get_screen (GTK_WINDOW (dialog->run_dialog));	
@@ -396,24 +391,30 @@ panel_run_dialog_execute (PanelRunDialog *dialog)
 		return;
 	}
 
-	url = gnome_vfs_make_uri_from_input_with_dirs (command,
-						       GNOME_VFS_MAKE_URI_DIR_HOMEDIR);
-	escaped = g_markup_escape_text (url, -1);
-	scheme = gnome_vfs_get_uri_scheme (url);
+	file = panel_util_get_file_optional_homedir (command);
+	uri = g_file_get_uri (file);
+	scheme = g_file_get_uri_scheme (file);
+	g_object_unref (file);
+
+	escaped = g_markup_escape_text (uri, -1);
 	result = FALSE;
 	
 	if (!g_ascii_strcasecmp (scheme, "http") ||
 	    !g_ascii_strcasecmp (scheme, "file"))
-		/* If this returns an http or file url, the url might refer to a
-		 * command that is somewhere in the path or an executable file.
-		 * So try executing it before displaying it. We execute the 
-		 * command in the user's shell so that it can do all the parameter
-		 * expansion and other magic for us.
-		 */
+		/* If this returns an http or file uri, the uri might refer to
+		 * a command that is somewhere in the path or an executable
+		 * file. So try executing it before displaying it. We execute
+		 * the command in the user's shell so that it can do all the
+		 * parameter expansion and other magic for us. */
 		result = panel_run_dialog_launch_command (dialog, disk, escaped);
 	
-	if (!result)
-		result = panel_run_dialog_show_url (dialog, url, escaped);
+	if (!result) {
+		GdkScreen *screen;
+		
+		screen = gtk_window_get_screen (GTK_WINDOW (dialog->run_dialog));
+		result = panel_show_uri (screen, uri,
+					 gtk_get_current_event_time (), NULL);
+	}
 		
 	if (result) {
 		/* only save working commands in history */
@@ -421,13 +422,14 @@ panel_run_dialog_execute (PanelRunDialog *dialog)
 					     TRUE, /* save item in history */
 					     command);
 		
-		/* only close the dialog if we successfully showed or launched something */
+		/* only close the dialog if we successfully showed or launched
+		 * something */
 		gtk_widget_destroy (dialog->run_dialog);
 	}
 
 	g_free (command);
 	g_free (disk);
-	g_free (url);
+	g_free (uri);
 	g_free (escaped);
 	g_free (scheme);
 }
@@ -601,7 +603,7 @@ panel_run_dialog_find_command_idle (PanelRunDialog *dialog)
 		if (path)
 			gtk_tree_path_free (path);
 		
-		panel_run_dialog_set_icon (dialog, NULL);
+		panel_run_dialog_set_icon (dialog, NULL, FALSE);
 	
 		dialog->find_command_idle_id = 0;
 		return FALSE;
@@ -632,9 +634,9 @@ panel_run_dialog_find_command_idle (PanelRunDialog *dialog)
 					    &iter,
 					    COLUMN_VISIBLE, TRUE,
 					    -1);
-		} else if (panel_util_utf8_strstrcase (exec, text) != NULL ||
-			   panel_util_utf8_strstrcase (name, text) != NULL ||
-			   panel_util_utf8_strstrcase (comment, text) != NULL) {
+		} else if (panel_g_utf8_strstrcase (exec, text) != NULL ||
+			   panel_g_utf8_strstrcase (name, text) != NULL ||
+			   panel_g_utf8_strstrcase (comment, text) != NULL) {
 			gtk_list_store_set (dialog->program_list_store,
 					    &iter,
 					    COLUMN_VISIBLE, TRUE,
@@ -660,7 +662,7 @@ panel_run_dialog_find_command_idle (PanelRunDialog *dialog)
 
 	gtk_tree_path_free (path);
 
-	panel_run_dialog_set_icon (dialog, found_icon);
+	panel_run_dialog_set_icon (dialog, found_icon, FALSE);
 	//FIXME update dialog->program_label
 
 	g_free (found_icon);
@@ -969,6 +971,7 @@ program_list_selection_changed (GtkTreeSelection *selection,
 	char         *temp;
 	char         *path, *stripped;
 	gboolean      terminal;
+	GKeyFile     *key_file;
 		
 	if (!gtk_tree_selection_get_selected (selection, &filter_model,
 					      &filter_iter))
@@ -983,65 +986,60 @@ program_list_selection_changed (GtkTreeSelection *selection,
 			    COLUMN_PATH, &path,
 			    -1);
 				  
-	if (path) {
-		GKeyFile *key_file;
+	if (!path)
+		return;
 
-		key_file = g_key_file_new ();
-
-		if (g_key_file_load_from_file (key_file, path,
-					       G_KEY_FILE_NONE, NULL)) {
-			dialog->use_program_list = TRUE;
-			if (dialog->desktop_path)
-				g_free (dialog->desktop_path);
-			dialog->desktop_path = g_strdup (path);
-			if (dialog->item_name)
-				g_free (dialog->item_name);
-			dialog->item_name = NULL;
-			
-			gtk_tree_model_foreach (GTK_TREE_MODEL (dialog->program_list_store),
-						panel_run_dialog_make_all_list_visible,
-						NULL);
-
-			/* Order is important here. We have to set the text
-			 * first so that the drag source is enabled, otherwise
-			 * the drag icon can't be set by
-			 * panel_run_dialog_set_icon.
-			 */
-			temp = panel_util_key_file_get_string (key_file,
-							       "Exec");
-			if (temp) {
-				stripped = remove_parameters (temp);
-				gtk_entry_set_text (GTK_ENTRY (dialog->gtk_entry), stripped);
-				g_free (stripped);
-			} else {
-				temp = panel_util_key_file_get_string (key_file,
-								       "URL");
-				gtk_entry_set_text (GTK_ENTRY (dialog->gtk_entry), sure_string (temp));
-			}
-			g_free (temp);
-
-			temp = panel_util_key_file_get_locale_string (key_file,
-								      "Icon");
-			panel_run_dialog_set_icon (dialog, temp);
-			g_free (temp);
-			
-			temp = panel_util_key_file_get_locale_string (key_file,
-								      "Comment");
-			//FIXME: if sure_string () == "", we should display "Will run..." as in entry_changed()
-			gtk_label_set_text (GTK_LABEL (dialog->program_label), sure_string (temp));
-			g_free (temp);
-
-			terminal = panel_util_key_file_get_boolean (key_file,
-								    "Terminal",
-								    FALSE);
-			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->terminal_checkbox),
-						      terminal);
-
-			g_key_file_free (key_file);
-                }
+	key_file = g_key_file_new ();
 
+	if (!g_key_file_load_from_file (key_file, path,
+					G_KEY_FILE_NONE, NULL)) {
+		g_key_file_free (key_file);
 		g_free (path);
-        }
+		return;
+	}
+
+	dialog->use_program_list = TRUE;
+	if (dialog->desktop_path)
+		g_free (dialog->desktop_path);
+	dialog->desktop_path = g_strdup (path);
+	if (dialog->item_name)
+		g_free (dialog->item_name);
+	dialog->item_name = NULL;
+	
+	/* Order is important here. We have to set the text first so that the
+	 * drag source is enabled, otherwise the drag icon can't be set by
+	 * panel_run_dialog_set_icon.
+	 */
+	temp = panel_key_file_get_string (key_file, "Exec");
+	if (temp) {
+		stripped = remove_parameters (temp);
+		gtk_entry_set_text (GTK_ENTRY (dialog->gtk_entry),
+				    stripped);
+		g_free (stripped);
+	} else {
+		temp = panel_key_file_get_string (key_file, "URL");
+		gtk_entry_set_text (GTK_ENTRY (dialog->gtk_entry),
+				    sure_string (temp));
+	}
+	g_free (temp);
+
+	temp = panel_key_file_get_locale_string (key_file, "Icon");
+	panel_run_dialog_set_icon (dialog, temp, FALSE);
+	g_free (temp);
+	
+	temp = panel_key_file_get_locale_string (key_file, "Comment");
+	//FIXME: if sure_string () == "", we should display "Will run..." as in entry_changed()
+	gtk_label_set_text (GTK_LABEL (dialog->program_label),
+			    sure_string (temp));
+	g_free (temp);
+
+	terminal = panel_key_file_get_boolean (key_file, "Terminal", FALSE);
+	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (dialog->terminal_checkbox),
+				      terminal);
+
+	g_key_file_free (key_file);
+
+	g_free (path);
 }
 
 static void
@@ -1655,7 +1653,7 @@ entry_drag_data_received (GtkEditable      *entry,
 		if (!uris [i] || !uris [i][0])
 			continue;
 		
-		file = gnome_vfs_get_local_path_from_uri (uris [i]);
+		file = g_filename_from_uri (uris [i], NULL, NULL);
 
 		/* FIXME: I assume the file is in utf8 encoding if coming from a URI? */
 		if (file) {
@@ -1712,24 +1710,24 @@ panel_run_dialog_create_desktop_file (PanelRunDialog *dialog)
 	GKeyFile   *key_file;
 	gboolean    exec = FALSE;
 	const char *text;
-	char       *uri;
+	GFile      *file;
 	char       *name;
 	char       *disk;
 	char       *scheme;
+	char       *save_uri;
 
 	text = gtk_entry_get_text (GTK_ENTRY (dialog->gtk_entry));
 	
 	if (!text || !text [0])
 		return NULL;
 		
-	key_file = panel_util_key_file_new_desktop ();
+	key_file = panel_key_file_new_desktop ();
 
 	disk = g_locale_from_utf8 (text, -1, NULL, NULL, NULL);
-	uri = gnome_vfs_make_uri_from_input_with_dirs (disk,
-						       GNOME_VFS_MAKE_URI_DIR_HOMEDIR);
+	file = panel_util_get_file_optional_homedir (disk);
 	g_free (disk);
 
-	scheme = gnome_vfs_get_uri_scheme (uri);
+	scheme = g_file_get_uri_scheme (file);
 	if (!g_ascii_strcasecmp (scheme, "http"))
 		exec = command_is_executable (text);
 	else if (!g_ascii_strcasecmp (scheme, "file"))
@@ -1737,43 +1735,48 @@ panel_run_dialog_create_desktop_file (PanelRunDialog *dialog)
 	g_free (scheme);
 		
 	if (exec) {
-		panel_util_key_file_set_string (key_file,
-						"Type", "Application");
-		panel_util_key_file_set_string (key_file,
-						"Exec", text);
+		panel_key_file_set_string (key_file, "Type", "Application");
+		panel_key_file_set_string (key_file, "Exec", text);
 		name = g_strdup (text);
 	} else {
-		panel_util_key_file_set_string (key_file,
-						"Type", "Link");
-		panel_util_key_file_set_string (key_file,
-						"URL", uri);
-		name = g_strdup (uri);
+		char *uri;
+
+		uri = g_file_get_uri (file);
+
+		panel_key_file_set_string (key_file, "Type", "Link");
+		panel_key_file_set_string (key_file, "URL", uri);
+		name = uri;
 	}
-	g_free (uri);
 
-	panel_util_key_file_set_locale_string (key_file, "Name",
-					       (dialog->item_name) ?
-					       	dialog->item_name : text);
+	g_object_unref (file);
 
-	panel_util_key_file_set_boolean (key_file, "Terminal",
-					 gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->terminal_checkbox)));
+	panel_key_file_set_locale_string (key_file, "Name",
+					  (dialog->item_name) ?
+					  dialog->item_name : text);
 
-	panel_util_key_file_set_locale_string (key_file, "Icon",
-					       dialog->icon_path);
+	panel_key_file_set_boolean (key_file, "Terminal",
+				    gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (dialog->terminal_checkbox)));
 
-	uri = panel_make_unique_desktop_uri (g_get_tmp_dir (), name);
-	disk = g_filename_from_uri (uri, NULL, NULL);
+	if (dialog->icon_path)
+		panel_key_file_set_locale_string (key_file, "Icon",
+						  dialog->icon_path);
+	else
+		panel_key_file_set_locale_string (key_file, "Icon",
+						  PANEL_ICON_LAUNCHER);
 
-	if (!disk || !panel_util_key_file_to_file (key_file, disk, NULL)) {
-		g_free (uri);
-		uri = NULL;
+	save_uri = panel_make_unique_desktop_uri (g_get_tmp_dir (), name);
+	disk = g_filename_from_uri (save_uri, NULL, NULL);
+
+	if (!disk || !panel_key_file_to_file (key_file, disk, NULL)) {
+		g_free (save_uri);
+		save_uri = NULL;
 	}
 
 	g_key_file_free (key_file);
 	g_free (disk);
 	g_free (name);
 
-	return uri;
+	return save_uri;
 }
 
 static void
@@ -1787,7 +1790,7 @@ pixmap_drag_data_get (GtkWidget          *run_dialog,
 	char *uri;
 
 	if (dialog->use_program_list && dialog->desktop_path)
-		uri = g_strdup (dialog->desktop_path);
+		uri = g_filename_to_uri (dialog->desktop_path, NULL, NULL);
 	else
 		uri = panel_run_dialog_create_desktop_file (dialog);
 
@@ -1799,12 +1802,47 @@ pixmap_drag_data_get (GtkWidget          *run_dialog,
 	}
 }
 
+static void
+panel_run_dialog_style_set (GtkWidget      *widget,
+			    GtkStyle       *prev_style,
+			    PanelRunDialog *dialog)
+{
+  if (dialog->icon_path) {
+	  char *icon_path;
+
+	  icon_path = g_strdup (dialog->icon_path);
+	  panel_run_dialog_set_icon (dialog, icon_path, TRUE);
+	  g_free (icon_path);
+  }
+}
+
+static void
+panel_run_dialog_screen_changed (GtkWidget      *widget,
+				 GdkScreen      *prev_screen,
+				 PanelRunDialog *dialog)
+{
+  if (dialog->icon_path) {
+	  char *icon_path;
+
+	  icon_path = g_strdup (dialog->icon_path);
+	  panel_run_dialog_set_icon (dialog, icon_path, TRUE);
+	  g_free (icon_path);
+  }
+}
+
 static void
 panel_run_dialog_setup_pixmap (PanelRunDialog *dialog,
 			       GladeXML       *gui)
 {
 	dialog->pixmap = glade_xml_get_widget (gui, "icon_pixmap");
 	
+	g_signal_connect (dialog->pixmap, "style-set",
+			  G_CALLBACK (panel_run_dialog_style_set),
+			  dialog);
+	g_signal_connect (dialog->pixmap, "screen-changed",
+			  G_CALLBACK (panel_run_dialog_screen_changed),
+			  dialog);
+
 	g_signal_connect (dialog->run_dialog, "drag_data_get",
 			  G_CALLBACK (pixmap_drag_data_get),
 			  dialog);
@@ -1856,6 +1894,17 @@ panel_run_dialog_new (GdkScreen *screen,
 	return dialog;
 }
 
+static void
+panel_run_dialog_disconnect_pixmap (PanelRunDialog *dialog)
+{
+	g_signal_handlers_disconnect_by_func (dialog->pixmap,
+			                      G_CALLBACK (panel_run_dialog_style_set),
+			                      dialog);
+	g_signal_handlers_disconnect_by_func (dialog->pixmap,
+			                      G_CALLBACK (panel_run_dialog_screen_changed),
+			                      dialog);
+}
+
 static void
 panel_run_dialog_static_dialog_destroyed (PanelRunDialog *dialog)
 {
